package com.pseudocode.photolib.transforms;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.imageio.ImageIO;
import javax.imageio.ImageReader;

/**
 * Image file formats
 */
public enum ImageFormat {
	/**
	 * Joint Photographic Experts Group (JPEG) Format
	 */
	JPEG {
		@Override
		public String extension() {
			return ".jpg";
		}

		@Override
		public String contentType() {
			return "image/jpeg";
		}
	},

	/**
	 * Portable Network Graphic (PNG) Format
	 */
	PNG {
		@Override
		public String extension() {
			return ".png";
		}

		@Override
		public String contentType() {
			return "image/png";
		}
	},

	/**
	 * Graphics Interchange Format (GIF)
	 */
	GIF {
		@Override
		public String extension() {
			return ".gif";
		}

		@Override
		public String contentType() {
			return "image/gif";
		}
	},

	/**
	 * Device Independent Bitmap (BMP) Format
	 */
	BMP {
		@Override
		public String extension() {
			return ".bmp";
		}

		@Override
		public String contentType() {
			return "image/bmp";
		}
	},

	/**
	 * Wireless Application Protocol Bitmap (WBMP) Format
	 */
	WBMP {
		@Override
		public String extension() {
			return ".wbmp";
		}

		@Override
		public String contentType() {
			return "image/vnd.wap.wbmp";
		}
	},

	/**
	 * Tagged Image File Format (TIFF)
	 */
	TIFF {
		@Override
		public String extension() {
			return ".tif";
		}

		@Override
		public String contentType() {
			return "image/tiff";
		}
	};

	private static final Map<String, ImageFormat> extensions = new HashMap<String, ImageFormat>();

	static {
		for (String ext : ImageIO.getReaderFileSuffixes()) {
			Iterator<ImageReader> readers = ImageIO.getImageReadersBySuffix(ext);
			while (readers.hasNext()) {
				try {
					String name = readers.next().getFormatName();
					if (name == null) {
						continue;
					}
					extensions.put('.'+ext, ImageFormat.valueOf(name.toUpperCase()));
					break;
				} catch (Exception ex) {}
			}
		}
	}

	public static Set<String> extensions() {
		return extensions.keySet();
	}

	public static ImageFormat byExtension(String ext) {
		if (ext == null) {
			return null;
		}
		if (ext.charAt(0) != '.') {
			ext = '.'+ext;
		}
		return extensions.get(ext);
	}
	
	public abstract String extension();
	
	public abstract String contentType();

	public static String normalizeExtension(String ext) {
		ImageFormat imageFormat = ImageFormat.byExtension(ext);
		if (imageFormat == null) {
			return ext;
		}

		return imageFormat.extension();
	}
}
