package com.pseudocode.photolib.transforms;

/**
 * Describes the operations which need to be performed to display the image.
 * All degrees are given as clockwise.
 */
public enum ImageOrientation {
	
	/* http://jpegclub.org/exif_orientation.html
	 * 
	 * EXIF orientation example from Adam M. Costello
	 * 
	 * Here is what the letter F would look like if it were tagged correctly and displayed
	 * by a program that ignores the orientation tag (thus showing the stored image):
	 *
	 *       1        2      3       4         5            6           7          8
	 * 
	 *  Phase #0
	 *              
	 *     888888  888888      88  88      8888888888  88                  88  8888888888
	 *     88          88      88  88      88  88      88  88          88  88      88  88
	 *     8888      8888    8888  8888    88          8888888888  8888888888          88
	 *     88          88      88  88
	 *     88          88  888888  888888
	 * 
	 *  Phase #1 (after applying operations once)
	 *              
	 *     888888  888888  888888  888888    888888      888888      888888      888888
	 *     88      88      88      88        88          88          88          88
	 *     8888    8888    8888    8888      8888        8888        8888        8888
	 *     88      88      88      88        88          88          88          88
	 *     88      88      88      88        88          88          88          88
	 * 
	 *  Phase #2 (after applying operations twice)
	 *     
	 *     888888  888888      88  88      8888888888  8888888888          88  88
	 *     88          88      88  88      88  88          88  88      88  88  88  88
	 *     8888      8888    8888  8888    88                  88  8888888888  8888888888
	 *     88          88      88  88
	 *     88          88  888888  888888
	 * 
	 *  Phase #3 (after applying operations three times)
	 *  
	 *     888888  888888  888888  888888    888888          88      888888          88
	 *     88      88      88      88        88              88      88              88
	 *     8888    8888    8888    8888      8888          8888      8888          8888
	 *     88      88      88      88        88              88      88              88
	 *     88      88      88      88        88          888888      88          888888
	 */

	/**
	 * Not specified
	 */
	INVALID,

	/**
	 * EXIF Orientation 1
	 * 0th row on TOP, 0th col on LEFT
	 */
	NORMAL,

	/**
	 * EXIF Orientation 2
	 * 0th row on TOP, 0th col on RIGHT
	 */
	FLIP_HORIZONTAL,

	/**
	 * EXIF Orientation 3
	 * 0th row on BOTTOM, 0th col on RIGHT
	 */
	ROTATE_180,

	/**
	 * EXIF Orientation 4
	 * 0th row on BOTTOM, 0th col on LEFT
	 */
	FLIP_VERTICAL,

	/**
	 * EXIF Orientation 5
	 * 0th row on LEFT, 0th col on TOP
	 */
	ROTATE_90_FLIP_HORIZONTAL,

	/**
	 * EXIF Orientation 6
	 * 0th row on RIGHT, 0th col on TOP
	 */
	ROTATE_90,

	/**
	 * EXIF Orientation 7
	 * 0th row on RIGHT, 0th col on BOTTOM
	 */
	ROTATE_270_FLIP_HORIZONTAL,

	/**
	 * EXIF Orientation 8
	 * 0th row on LEFT, 0th col on BOTTOM
	 */
	ROTATE_270;

	/**
	 * Converts the EXIF int value into the proper enum equivalent
	 * @param value
	 * @return
	 */
	public static ImageOrientation fromEXIF(int value) {
		try {
			return ImageOrientation.values()[value];

		} catch (Exception ex) {
			return ImageOrientation.INVALID;
		}
	}
}
