package com.pseudocode.photolib.transforms;

import java.awt.AlphaComposite;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.awt.image.ColorConvertOp;
import java.awt.image.ConvolveOp;
import java.awt.image.Kernel;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;

import javax.imageio.IIOImage;
import javax.imageio.ImageIO;
import javax.imageio.ImageWriteParam;
import javax.imageio.ImageWriter;
import javax.imageio.stream.FileImageOutputStream;
import javax.imageio.stream.ImageOutputStream;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class ImageTransformer {
	private static final Logger log = LoggerFactory.getLogger(ImageTransformer.class);

	/**
	 * Ensures specified transforms exist in targetDir otherwise generates.
	 * @param sourceFile original image
	 * @param targetDir transform directory
	 * @param orientation camera orientation of image (to be removed before applying transformations)
	 * @param overwrite if should overwrite existing transforms
	 * @param transforms transform descriptors
	 */
	public static void applyTransforms(
			File sourceFile,
			File targetDir,
			ImageOrientation orientation,
			boolean overwrite,
			ImageTransform... transforms) {

		if (sourceFile == null) {
			throw new NullPointerException("sourceFile");
		}
		if (targetDir == null) {
			throw new NullPointerException("targetDir");
		}
		if (!sourceFile.exists()) {
			throw new IllegalArgumentException("File does not exist: "+sourceFile.getPath());
		}
		if (targetDir.exists() && !targetDir.isDirectory()) {
			throw new IllegalArgumentException("targetDir must be a directory");
		}
		if (transforms == null || transforms.length < 1) {
			return;
		}

		BufferedImage originalImage = null;

		for (ImageTransform transform : transforms) {
			ImageFormat imageFormat = (transform.imageFormat() != null) ?
				transform.imageFormat() :
				ImageFormat.byExtension( getExtension(sourceFile.getName()) );

			File targetFile = getFile(targetDir, transform, imageFormat);

			// check if transform already exists
			if (!overwrite && targetFile.exists()) {
				log.trace(sourceFile.getPath()+" => "+targetFile.getName()+" exists");
				continue;
			}

			if (originalImage == null) {
				long nano = System.nanoTime();
				try {
					// lazy load image data
					originalImage = loadImage(sourceFile, orientation);

				} catch (IOException ex) {
					log.error("ERROR loading image: "+ex.getMessage(), ex);
				}

				if (originalImage == null) {
					// use dummy image
					originalImage = new BufferedImage(100, 100, BufferedImage.TYPE_INT_RGB);
				}

				nano = System.nanoTime() - nano;
				log.info(sourceFile.getName()+" loaded ("+(nano/1e6)+"ms)");
			}

			long nano = System.nanoTime();

			BufferedImage transformedImage = applyTransform(originalImage, transform);

			try {
				saveImage(transformedImage, targetFile, imageFormat, transform.quality());

			} catch (IOException ex) {
				log.error("ERROR saving image: "+ex.getMessage(), ex);
			}

			// free up some resources
			transformedImage.flush();

			nano = System.nanoTime() - nano;
			log.info(sourceFile.getName()+" => "+targetFile.getName()+" transformed ("+(nano/1e6)+"ms)");
		}

		// free up some resources
		if (originalImage != null) {
			originalImage.flush();
		}
	}

	/**
	 * Generates specified transform in targetDir overwriting any existing.
	 * @param sourceFile original image
	 * @param targetFile transform image
	 * @param orientation camera orientation of image (to be removed before applying transformations)
	 * @param overwrite if should overwrite existing transform
	 * @param transform transform descriptors
	 */
	public static void applyTransform(
			File sourceFile,
			File targetFile,
			ImageOrientation orientation,
			boolean overwrite,
			ImageTransform transform) {

		if (sourceFile == null) {
			throw new NullPointerException("sourceFile");
		}
		if (targetFile == null) {
			throw new NullPointerException("targetFile");
		}
		if (!sourceFile.exists()) {
			throw new IllegalArgumentException("File does not exist: "+sourceFile.getPath());
		}
		if (targetFile.exists() && targetFile.isDirectory()) {
			throw new IllegalArgumentException("targetFile is a directory");
		}
		if (transform == null) {
			return;
		}

		// check if transform already exists
		if (!overwrite && targetFile.exists()) {
			log.trace(sourceFile.getPath()+" => "+targetFile.getName()+" exists");
			return;
		}

		ImageFormat imageFormat = (transform.imageFormat() != null) ?
			transform.imageFormat() :
			ImageFormat.byExtension( getExtension(sourceFile.getName()) );

		BufferedImage originalImage = null;
		try {
			// load image data
			originalImage = loadImage(sourceFile, orientation);

		} catch (IOException ex) {
			log.error("ERROR loading image: "+ex.getMessage(), ex);
		}

		if (originalImage == null) {
			// use dummy image
			originalImage = new BufferedImage(100, 100, BufferedImage.TYPE_INT_RGB);
		}

		long nano = System.nanoTime();

		BufferedImage transformedImage = applyTransform(originalImage, transform);

		try {
			saveImage(transformedImage, targetFile, imageFormat, transform.quality());

		} catch (IOException ex) {
			log.error("ERROR saving image: "+ex.getMessage(), ex);
		}

		// free up some resources
		transformedImage.flush();

		nano = System.nanoTime() - nano;
		log.info(sourceFile.getName()+" => "+targetFile.getName()+" transformed ("+(nano/1e6)+"ms)");

		// free up some resources
		originalImage.flush();
	}

	/**
	 * Applies the transformation to the loaded image
	 * @param image
	 * @param transform
	 * @return
	 */
	public static BufferedImage applyTransform(
			BufferedImage image,
			ImageTransform transform) {

		// apply transformation rotation
		image = rotate(image, transform.orientation());

		// determine metrics for this original / transform combination
		ImageShape shape = calcShape(transform, image.getWidth(), image.getHeight());

		// scale and crop
		image = resize(image, shape);

		// apply transformation blur
		image = gaussianBlur(image, transform.blur());

		return image;
	}
	
	public static BufferedImage rotate(BufferedImage original, ImageOrientation orientation) {
		if (orientation == null) {
			return original;
		}

		AffineTransform affine;
		int temp,
			width = original.getWidth(),
			height = original.getHeight();

		switch (orientation) {
			case FLIP_HORIZONTAL:// EXIF 2
				affine = new AffineTransform();
				affine.translate(width, 0);
				affine.scale(-1.0, 1.0);
				break;

			case ROTATE_180:// EXIF 3
				affine = new AffineTransform();
				affine.translate(width, height);
				affine.quadrantRotate(2);
				break;

			case FLIP_VERTICAL:// EXIF 4
				affine = new AffineTransform();
				affine.translate(0, height);
				affine.scale(1.0, -1.0);
				break;

			case ROTATE_90_FLIP_HORIZONTAL:// EXIF 5
				// swap for sideways rotations
				temp = width;
				width = height;
				height = temp;

				// rotate
				affine = new AffineTransform();
				affine.translate(width, 0);
				affine.quadrantRotate(1);

				// flip (axes haven't rotated yet)
				affine.translate(0, width);
				affine.scale(1.0, -1.0);
				break;

			case ROTATE_90:// EXIF 6
				// swap for sideways rotations
				temp = width;
				width = height;
				height = temp;

				affine = new AffineTransform();
				affine.translate(width, 0);
				affine.quadrantRotate(1);
				break;

			case ROTATE_270_FLIP_HORIZONTAL:// EXIF 7
				// swap for sideways rotations
				temp = width;
				width = height;
				height = temp;

				// rotate
				affine = new AffineTransform();
				affine.translate(0, height);
				affine.quadrantRotate(-1);

				// flip (axes haven't rotated yet)
				affine.translate(0, width);
				affine.scale(1.0, -1.0);
				break;

			case ROTATE_270:// EXIF 8
				// swap for sideways rotations
				temp = width;
				width = height;
				height = temp;

				affine = new AffineTransform();
				affine.translate(0, height);
				affine.quadrantRotate(-1);
				break;

			default:
				// NOOP for normal or unspecified
				return original;
		}
		
		boolean preserveAlpha = original.getColorModel().hasAlpha();
		int imageType = preserveAlpha ? BufferedImage.TYPE_INT_ARGB : BufferedImage.TYPE_INT_RGB;

		BufferedImage image = new BufferedImage(width, height, imageType);

		Graphics2D g = image.createGraphics();
		try {
			g.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);
			g.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION, RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);
			g.setRenderingHint(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_QUALITY);
			g.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
			g.setRenderingHint(RenderingHints.KEY_DITHERING, RenderingHints.VALUE_DITHER_ENABLE);
			if (preserveAlpha) {
				g.setComposite(AlphaComposite.Src);
			}
			g.drawImage(original, affine, null);

		} finally {
			g.dispose();
		}

		return image;
	}

	/**
	 * Incorporates multi-step scaling in powers of two for better results with bicubic interpolation.
	 */
	private static BufferedImage resize(BufferedImage original, ImageShape shape) {
		final boolean preserveAlpha = original.getColorModel().hasAlpha();
		final int imageType = preserveAlpha ? BufferedImage.TYPE_INT_ARGB : BufferedImage.TYPE_INT_RGB;

		// adjust color space if needed
		BufferedImage image = original;
		Graphics2D g;

		// begin scale down at a higher multiple of two
		// then halve until the final scale / crop operation
		// scale up skips this loop since it has no multiplier
		for (int m=shape.multiplier(); m>0; m--) {
			BufferedImage next = new BufferedImage(shape.width()<<m, shape.height()<<m, imageType);
			g = next.createGraphics();
			try {
				if (m != shape.multiplier()) {
					// NOTE: if exact halving we use BILINEAR, as BICUBIC increases acutance giving an over-sharpened look
					g.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);
				} else {
					g.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);
				}
				g.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION, RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);
				g.setRenderingHint(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_QUALITY);
				g.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
				if (preserveAlpha) {
					g.setComposite(AlphaComposite.Src);
				}
				g.drawImage(image, 0, 0, shape.width()<<m, shape.height()<<m, null);
				image = next;

			} finally {
				g.dispose();
			}

			log.trace("rendering at: "+(shape.width()<<m)+'x'+(shape.height()<<m));
		}

		BufferedImage next = shape.crop() ?
			new BufferedImage(shape.right(), shape.bottom(), imageType) :
			new BufferedImage(shape.width(), shape.height(), imageType);

		g = next.createGraphics();
		try {
			if (shape.multiplier() > 0) {
				// NOTE: if exact halving we use BILINEAR, as BICUBIC increases acutance giving an over-sharpened look
				g.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);
			} else {
				g.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BICUBIC);
			}
			g.setRenderingHint(RenderingHints.KEY_ALPHA_INTERPOLATION, RenderingHints.VALUE_ALPHA_INTERPOLATION_QUALITY);
			g.setRenderingHint(RenderingHints.KEY_COLOR_RENDERING, RenderingHints.VALUE_COLOR_RENDER_QUALITY);
			g.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
			if (preserveAlpha) {
				g.setComposite(AlphaComposite.Src);
			}
			if (shape.crop()) {
				g.drawImage(image, shape.left(), shape.top(), shape.width(), shape.height(), null);
			} else {
				g.drawImage(image, 0, 0, shape.width(), shape.height(), null);
			}
			image = next;

		} finally {
			g.dispose();
		}

		log.trace("rendered at: "+shape.width()+'x'+shape.height());
		return image;
	}

	/**
	 * Performs two 1-D Gaussian Blur operations
	 * @param image
	 * @param radius
	 * @return
	 */
	public static BufferedImage gaussianBlur(BufferedImage image, int radius) {
		final boolean preserveAlpha = image.getColorModel().hasAlpha();
		final int imageType = preserveAlpha ? BufferedImage.TYPE_INT_ARGB : BufferedImage.TYPE_INT_RGB;

		if (radius > 0) {
			if (image.getType() != imageType) {
				// adjust color space if needed
				BufferedImage next = new BufferedImage(image.getWidth(), image.getHeight(), imageType);
				image = new ColorConvertOp(null).filter(image, next);
			}

			// soften the final halving via two 1D Gaussian blur filters
			image = getGaussianBlurFilter(radius, false).filter(image, null);
			image = getGaussianBlurFilter(radius, true).filter(image, null);
		}
		return image;
	}

	/**
	 * Generates convolve kernel for Gaussian Blur
	 * @param radius
	 * @param horizontal
	 * @return
	 */
	private static ConvolveOp getGaussianBlurFilter(int radius, boolean horizontal) {
		if (radius < 1) {
			throw new IllegalArgumentException("Radius must be positive");
		}

		int size = radius * 2 + 1;
		float[] data = new float[size];

		float sigma = radius / 3.0f;
		float twoSigmaSquare = 2.0f * sigma * sigma;
		float sigmaRoot = (float)Math.sqrt(twoSigmaSquare * Math.PI);
		float total = 0.0f;

		for (int i = -radius; i <= radius; i++) {
			float distance = i * i;
			int index = i + radius;
			data[index] = (float) Math.exp(-distance / twoSigmaSquare) / sigmaRoot;
			total += data[index];
		}

		for (int i = 0; i < data.length; i++) {
			data[i] /= total;
		}		 

		Kernel kernel;
		if (horizontal) {
			kernel = new Kernel(size, 1, data);
		} else {
			kernel = new Kernel(1, size, data);
		}
		return new ConvolveOp(kernel, ConvolveOp.EDGE_NO_OP, null);
	}

	/**
	 * Calculates the render shape resulting from applying the transform to image dimensions
	 * @param transform
	 * @param width
	 * @param height
	 * @return
	 */
	private static ImageShape calcShape(ImageTransform transform, int width, int height) {
		// use original dimensions if transform doesn't constrain
		int transformWidth = (transform.maxWidth() > 0) ? transform.maxWidth() : width;
		int transformHeight = (transform.maxHeight() > 0) ? transform.maxHeight() : height;

		if (!transform.oversize()) {
			// prevent transform scaling larger than original
			transformWidth = Math.min(transformWidth, width);
			transformHeight = Math.min(transformHeight, height);
		}

		// portrait < 1.0, square == 1.0, landscape > 1.0
		double transformRatio = (double)transformWidth / (double)transformHeight;
		double imageRatio = (double)width / (double)height;

		// determine constraint axis
		boolean constrainHeight = transform.oversize() ?
				(transformRatio < imageRatio) :
				(transformRatio >= imageRatio);

		ImageShape shape;

		if (constrainHeight) {
			shape = new ImageShape()
				.height(transformHeight)
				.width((int)Math.round((double)transformHeight * imageRatio));

			if (transform.crop()) {
				shape
					.crop(true)
					.left((int)Math.round(transform.horzAlign() * ((double)transformWidth - (double)shape.width())))
					.right(transformWidth)
					.bottom(transformHeight);
			}

			shape.multiplier(calcMultiplier(height, transformHeight));

		} else {
			shape = new ImageShape()
				.height((int)Math.round((double)transformWidth / imageRatio))
				.width(transformWidth);

			if (transform.crop()) {
				shape
					.crop(true)
					.top((int)Math.round(transform.vertAlign() * ((double)transformHeight - (double)shape.height())))
					.right(transformWidth)
					.bottom(transformHeight);
			}

			shape.multiplier(calcMultiplier(width, transformWidth));
		}

		return shape;
	}

	/**
	 * calc scaling down hints for better anti-aliasing
	 * find the greater power of two to start scaling
	 * then halve the image until is the transform size
	 * @param originalDim
	 * @param transformDim
	 * @return
	 */
	private static int calcMultiplier(int originalDim, int transformDim) {
		log.trace("transformDim: "+transformDim+"; originalDim: "+originalDim);
		double power = log2((double)originalDim / (double)transformDim);
		int multiplier = (int)Math.min(Math.floor(power)-1, 4);
		log.trace("multiplier: "+multiplier);
		return multiplier;
	}

	/**
	 * Generates the transformation file name
	 * @param transform the image transform
	 * @param imageFormat the image format
	 * @return
	 */
	private static File getFile(File targetDir, ImageTransform transform, ImageFormat imageFormat) {
		// ensure valid values
		if (targetDir == null) {
			throw new NullPointerException("targetDir");
		}
		if (transform == null) {
			throw new NullPointerException("transform");
		}
		if (imageFormat == null) {
			throw new NullPointerException("imageFormat");
		}

		//"{target-dir}/{transform-name}.{format}"
		return new File(targetDir, 
			transform.toString( new StringBuilder(20) )
			.append(imageFormat.extension())
			.toString());
	}

	/**
	 * Log base 2
	 * @param value
	 * @return
	 */
	private static double log2(double value) {
		return Math.log(value) / Math.log(2.0);
	}

	/**
	 * Loads the original image data
	 * @param sourceFile
	 * @param orientation 
	 * @return
	 * @throws IOException 
	 */
	public static BufferedImage loadImage(
			File sourceFile,
			ImageOrientation orientation) throws IOException {

		if (sourceFile == null) {
			throw new NullPointerException("sourceFile");
		}

		BufferedImage image = null;
		InputStream stream = null;
		try {
			stream = new FileInputStream(sourceFile);

			// load source image data
			image = ImageIO.read(stream);

		} finally {
			if (stream != null) {
				try {
					stream.close();
				} catch (IOException ex) {
					log.error("ERROR closing image input stream: "+ex.getMessage(), ex);
				}
			}
		}

		if (image != null) {
			// undo any camera orientation before transformation
			image = rotate(image, orientation);
		}
		return image;
	}

	/**
	 * Loads the original image data
	 * @param stream
	 * @param orientation 
	 * @return
	 * @throws IOException 
	 */
	public static BufferedImage loadImage(
			InputStream stream,
			ImageOrientation orientation) throws IOException {

		if (stream == null) {
			throw new NullPointerException("stream");
		}

		// load source image data
		BufferedImage image = ImageIO.read(stream);

		if (image != null) {
			// undo any camera orientation before transformation
			image = rotate(image, orientation);
		}
		return image;
	}

	/**
	 * Saves the image to the target file
	 * @param image
	 * @param targetFile
	 * @param imageFormat
	 * @param quality
	 * @throws IOException
	 */
	public static void saveImage(
			BufferedImage image,
			File targetFile,
			ImageFormat imageFormat,
			int quality) throws IOException {

		if (image == null) {
			throw new NullPointerException("image");
		}
		if (targetFile == null) {
			throw new NullPointerException("targetFile");
		}

		// prep save path
		if (!targetFile.getParentFile().exists()) {
			targetFile.getParentFile().mkdirs();
		}

		ImageOutputStream stream = null;
		try {
			stream = new FileImageOutputStream(targetFile);
			saveImage(image, stream, imageFormat, quality);

		} finally {
			if (stream != null) {
				try {
					stream.close();
				} catch (IOException ex) {
					log.error("ERROR closing tranform stream: "+ex.getMessage(), ex);
				}
			}
		}
	}

	/**
	 * Saves the image to the target file
	 * @param image
	 * @param targetFile
	 * @param imageFormat
	 * @param quality
	 * @throws IOException
	 */
	public static void saveImage(
			BufferedImage image,
			ImageOutputStream stream,
			ImageFormat imageFormat,
			int quality) throws IOException {

		if (image == null) {
			throw new NullPointerException("image");
		}
		if (stream == null) {
			throw new NullPointerException("stream");
		}
		if (imageFormat == null) {
			imageFormat = ImageFormat.JPEG;
		}

		if (imageFormat != ImageFormat.JPEG || quality <= 0 || quality > 100) {
			// accept codec defaults
			ImageIO.write(image, imageFormat.name(), stream);

		} else {
			ImageWriter writer = ImageIO.getImageWritersByFormatName(imageFormat.name()).next();
			ImageWriteParam param = writer.getDefaultWriteParam();
			param.setCompressionMode(ImageWriteParam.MODE_EXPLICIT);
			param.setCompressionQuality(quality / 100f);
			writer.setOutput(stream);
			writer.write(null, new IIOImage(image, null, null), param);
		}
	}

	private static String getExtension(String name) {
		if (name == null) {
			return "";
		}

		int dot = name.lastIndexOf('.');
		if (dot < 0) {
			return "";
		}

		return name.substring(dot).toLowerCase();
	}
}
